@@ -11,14 +11,13 @@
 /*  This is a new version using the MiGS graphics subsystem for             */
 /*  blitting and display.                                                   */
 /*                                                                          */
-/*  Press F1 when running this program to have a list of key-bindings       */
+/*  Press ? when running this program to have a list of key-bindings.       */
 /*                                                                          */
 /****************************************************************************/
 
 
 #include "ftcommon.h"
 #include "common.h"
-#include <math.h>
 #include <stdio.h>
 
   /* the following header shouldn't be used in normal programs */
@@ -99,8 +98,11 @@
     double         radius;
     double         slant;
 
-    int            cff_hinting_engine;
-    int            tt_interpreter_version;
+    unsigned int   cff_hinting_engine;
+    unsigned int   tt_interpreter_versions[3];
+    int            num_tt_interpreter_versions;
+    int            tt_interpreter_version_idx;
+    FT_Bool        warping;
 
     int            font_idx;
     int            offset;            /* as selected by the user */
@@ -108,16 +110,16 @@
     int            num_fails;
     int            preload;
 
-    int            use_custom_lcd_filter;
+    int            lcd_filter;
     unsigned char  filter_weights[5];
     int            fw_idx;
 
   } status = { 1,
                DIM_X, DIM_Y, RENDER_MODE_ALL, FT_ENCODING_NONE,
-               72, 48, -1, 1.0, 0.04, 0.04, 0.02, 0.22,
-               0, 0, /* default values are set at runtime */
+               72, 48, -1, GAMMA, 0.04, 0.04, 0.02, 0.22,
+               0, { 0 }, 0, 0, 0, /* default values are set at runtime */
                0, 0, 0, 0, 0,
-               0, { 0x10, 0x40, 0x70, 0x40, 0x10 }, 2 };
+               FT_LCD_FILTER_DEFAULT, { 0x08, 0x4D, 0x56, 0x4D, 0x08 }, 2 };
 
 
   static FTDemo_Display*  display;
@@ -182,7 +184,8 @@
     face = size->face;
     slot = face->glyph;
 
-    radius = status.radius * ( status.ptsize * status.res / 72 );
+    radius = (FT_Fixed)(
+               status.radius * ( status.ptsize * status.res / 72 ) );
 
     FT_Stroker_Set( handle->stroker, radius,
                     FT_STROKER_LINECAP_ROUND,
@@ -193,13 +196,13 @@
 
     for ( i = offset; i < num_indices; i++ )
     {
-      int  glyph_idx;
+      FT_UInt  glyph_idx;
 
 
       if ( handle->encoding == FT_ENCODING_NONE )
-        glyph_idx = i;
+        glyph_idx = (FT_UInt)i;
       else
-        glyph_idx = FTDemo_Get_Index( handle, i );
+        glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)i );
 
       error = FT_Load_Glyph( face, glyph_idx,
                              handle->load_flags | FT_LOAD_NO_BITMAP );
@@ -301,13 +304,13 @@
 
     for ( i = offset; i < num_indices; i++ )
     {
-      int  glyph_idx;
+      FT_UInt  glyph_idx;
 
 
       if ( handle->encoding == FT_ENCODING_NONE )
-        glyph_idx = i;
+        glyph_idx = (FT_UInt)i;
       else
-        glyph_idx = FTDemo_Get_Index( handle, i );
+        glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)i );
 
       error = FT_Load_Glyph( face, glyph_idx, handle->load_flags );
       if ( !error )
@@ -371,20 +374,20 @@
     slot = face->glyph;
 
     ystr = status.ptsize * status.res / 72;
-    xstr = status.xbold_factor * ystr;
-    ystr = status.ybold_factor * ystr;
+    xstr = (FT_Pos)( status.xbold_factor * ystr );
+    ystr = (FT_Pos)( status.ybold_factor * ystr );
 
     have_topleft = 0;
 
     for ( i = offset; i < num_indices; i++ )
     {
-      int  glyph_idx;
+      FT_UInt  glyph_idx;
 
 
       if ( handle->encoding == FT_ENCODING_NONE )
-        glyph_idx = i;
+        glyph_idx = (FT_UInt)i;
       else
-        glyph_idx = FTDemo_Get_Index( handle, i );
+        glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)i );
 
       error = FT_Load_Glyph( face, glyph_idx, handle->load_flags );
       if ( !error )
@@ -482,13 +485,13 @@
 
     for ( i = offset; i < num_indices; i++ )
     {
-      int  glyph_idx;
+      FT_UInt  glyph_idx;
 
 
       if ( handle->encoding == FT_ENCODING_NONE )
-        glyph_idx = i;
+        glyph_idx = (FT_UInt)i;
       else
-        glyph_idx = FTDemo_Get_Index( handle, i );
+        glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)i );
 
       error = FTDemo_Draw_Index( handle, display, glyph_idx, &x, &y );
 
@@ -564,9 +567,13 @@
       {
         p  = Text;
         ch = utf8_next( &p, pEnd );
+
+        /* not a single character of the text string could be displayed */
+        if ( !have_topleft )
+          return error;
       }
 
-      glyph_idx = FTDemo_Get_Index( handle, ch );
+      glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)ch );
 
       error = FTDemo_Draw_Index( handle, display, glyph_idx, &x, &y );
 
@@ -681,7 +688,7 @@
       }
 
       start = snprintf( text, 256, "%g: ", pt_size / 64.0 );
-      snprintf( text + start, 256 - start, "%s", p );
+      snprintf( text + start, (unsigned int)( 256 - start ), "%s", p );
 
       p    = text;
       pEnd = p + strlen( text );
@@ -699,9 +706,13 @@
           p    = Text;
           oldp = p;
           ch   = utf8_next( &p, pEnd );
+
+          /* not a single character of the text string could be displayed */
+          if ( !have_topleft )
+            return error;
         }
 
-        glyph_idx = FTDemo_Get_Index( handle, ch );
+        glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)ch );
 
         error = FTDemo_Draw_Index( handle, display, glyph_idx, &x, &y );
 
@@ -795,7 +806,7 @@
     grWriteln( "Up, Down    adjust size by 1 unit       r, R        adjust stroking radius  " );
     grWriteln( "PgUp, PgDn  adjust size by 10 units                  (in mode 4)            " );
     grWriteln( "                                                                            " );
-    grWriteln( "Left, Right adjust index by 1           L           toggle custom           " );
+    grWriteln( "Left, Right adjust index by 1           L           cycle through           " );
     grWriteln( "F7, F8      adjust index by 10                       LCD filtering          " );
     grWriteln( "F9, F10     adjust index by 100         [, ]        select custom LCD       " );
     grWriteln( "F11, F12    adjust index by 1000                      filter weight         " );
@@ -803,10 +814,12 @@
     grWriteln( "h           toggle hinting              -, +(=)     adjust selected custom  " );
     grWriteln( "H           cycle through hinting                    LCD filter weight      " );
     grWriteln( "             engines (if available)                                         " );
-    grWriteln( "f           toggle forced auto-         G           show gamma ramp         " );
-    grWriteln( "             hinting (if hinting)       g, v        adjust gamma value      " );
+    grWriteln( "f           toggle forced auto-         g, v        adjust gamma value      " );
+    grWriteln( "             hinting (if hinting)                                           " );
+    grWriteln( "w           toggle warping (in light                                        " );
+    grWriteln( "             AA mode, if available)     q, ESC      quit ftview             " );
     grWriteln( "                                                                            " );
-    grWriteln( "a           toggle anti-aliasing        q, ESC      quit ftview             " );
+    grWriteln( "a           toggle anti-aliasing                                            " );
     /*          |----------------------------------|    |----------------------------------| */
     grLn();
     grLn();
@@ -817,78 +830,10 @@
   }
 
 
-  static void
-  event_gamma_grid( void )
-  {
-    grEvent  dummy_event;
-    int      g;
-    int      yside  = 11;
-    int      xside  = 10;
-    int      levels = 17;
-    int      gammas = 30;
-    int      x_0    = ( display->bitmap->width - levels * xside ) / 2;
-    int      y_0    = ( display->bitmap->rows - gammas * ( yside + 1 ) ) / 2;
-    int      pitch  = display->bitmap->pitch;
-
-
-    FTDemo_Display_Clear( display );
-    grGotobitmap( display->bitmap );
-
-    if ( pitch < 0 )
-      pitch = -pitch;
-
-    memset( display->bitmap->buffer, 100, pitch * display->bitmap->rows );
-
-    grWriteCellString( display->bitmap, 0, 0, "Gamma grid",
-                       display->fore_color );
-
-    for ( g = 1; g <= gammas; g++ )
-    {
-      double  ggamma = 0.1 * g;
-      char    temp[6];
-      int     y = y_0 + ( yside + 1 ) * ( g - 1 );
-      int     nx, ny;
-
-      unsigned char*  line = display->bitmap->buffer +
-                             y * display->bitmap->pitch;
-
-
-      if ( display->bitmap->pitch < 0 )
-        line -= display->bitmap->pitch * ( display->bitmap->rows - 1 );
-
-      line += x_0 * 3;
-
-      grSetPixelMargin( x_0 - 32, y + ( yside - 8 ) / 2 );
-      grGotoxy( 0, 0 );
-
-      sprintf( temp, "%.1f", ggamma );
-      grWrite( temp );
-
-      for ( ny = 0; ny < yside; ny++, line += display->bitmap->pitch )
-      {
-        unsigned char*  dst = line;
-
-
-        for ( nx = 0; nx < levels; nx++, dst += 3 * xside )
-        {
-          double  p   = nx / (double)( levels - 1 );
-          int     gm  = (int)( 255.0 * pow( p, ggamma ) );
-
-
-          memset( dst, gm, xside * 3 );
-        }
-      }
-    }
-
-    grRefreshSurface( display->surface );
-    grListenSurface( display->surface, gr_event_key, &dummy_event );
-  }
-
-
   static int
-  event_cff_hinting_engine_change( int  delta )
+  event_cff_hinting_engine_change( unsigned int  delta )
   {
-    int  new_cff_hinting_engine;
+    unsigned int  new_cff_hinting_engine = 0;
 
 
     if ( delta )
@@ -919,18 +864,14 @@
   static int
   event_tt_interpreter_version_change( void )
   {
-    FT_UInt  new_interpreter_version;
-
-
-    if ( status.tt_interpreter_version == TT_INTERPRETER_VERSION_35 )
-      new_interpreter_version = TT_INTERPRETER_VERSION_38;
-    else
-      new_interpreter_version = TT_INTERPRETER_VERSION_35;
+    status.tt_interpreter_version_idx += 1;
+    status.tt_interpreter_version_idx %= status.num_tt_interpreter_versions;
 
     error = FT_Property_Set( handle->library,
                              "truetype",
                              "interpreter-version",
-                             &new_interpreter_version );
+                             &status.tt_interpreter_versions[
+                               status.tt_interpreter_version_idx] );
 
     if ( !error )
     {
@@ -938,7 +879,6 @@
       /* lazy to walk over all loaded fonts to check whether they */
       /* are of type TTF, then unloading them explicitly.         */
       FTC_Manager_Reset( handle->cache_manager );
-      status.tt_interpreter_version = new_interpreter_version;
       return 1;
     }
 
@@ -946,6 +886,34 @@
   }
 
 
+  static int
+  event_warping_change( void )
+  {
+    if ( handle->lcd_mode == LCD_MODE_LIGHT )
+    {
+      FT_Bool  new_warping_state = !status.warping;
+
+
+      error = FT_Property_Set( handle->library,
+                               "autofitter",
+                               "warping",
+                               &new_warping_state );
+
+      if ( !error )
+      {
+        /* Resetting the cache is perhaps a bit harsh, but I'm too  */
+        /* lazy to walk over all loaded fonts to check whether they */
+        /* are auto-hinted, then unloading them explicitly.         */
+        FTC_Manager_Reset( handle->cache_manager );
+        status.warping = new_warping_state;
+        return 1;
+      }
+    }
+
+    return 0;
+  }
+
+
   static void
   event_gamma_change( double  delta )
   {
@@ -1160,7 +1128,7 @@
       break;
 
     case grKEY( 'f' ):
-      if ( handle->hinted && handle->lcd_mode != LCD_MODE_LIGHT )
+      if ( handle->hinted )
       {
         handle->autohint = !handle->autohint;
         FTDemo_Update_Current_Flags( handle );
@@ -1212,6 +1180,10 @@
       }
       break;
 
+    case grKEY( 'w' ):
+      status.update = event_warping_change();
+      break;
+
     case grKeySpace:
       event_render_mode_change( 1 );
       status.update = 1;
@@ -1222,11 +1194,6 @@
       status.update = 1;
       break;
 
-    case grKEY( 'G' ):
-      event_gamma_grid();
-      status.update = 1;
-      break;
-
     case grKEY( 's' ):
       if ( status.render_mode == RENDER_MODE_SLANTED )
         status.update = event_slant_change( 0.02 );
@@ -1327,18 +1294,28 @@
       FTC_Manager_RemoveFaceID( handle->cache_manager,
                                 handle->scaler.face_id );
 
-      status.use_custom_lcd_filter = !status.use_custom_lcd_filter;
-      if ( status.use_custom_lcd_filter )
-        FT_Library_SetLcdFilterWeights( handle->library,
-                                        status.filter_weights );
-      else
-        FT_Library_SetLcdFilterWeights( handle->library,
-                                        (unsigned char*)"\x10\x40\x70\x40\x10" );
+      status.lcd_filter++;
+      switch ( status.lcd_filter )
+      {
+        case FT_LCD_FILTER_NONE:
+        case FT_LCD_FILTER_DEFAULT:
+        case FT_LCD_FILTER_LIGHT:
+        case FT_LCD_FILTER_LEGACY1:
+          FT_Library_SetLcdFilter( handle->library,
+                                   (FT_LcdFilter)status.lcd_filter );
+          break;
+        default:
+          FT_Library_SetLcdFilter( handle->library, FT_LCD_FILTER_DEFAULT );
+          FT_Library_SetLcdFilterWeights( handle->library,
+                                          status.filter_weights );
+          status.lcd_filter = -1;
+      }
+
       status.update = 1;
       break;
 
     case grKEY( '[' ):
-      if ( status.use_custom_lcd_filter )
+      if ( status.lcd_filter < 0 )
       {
         status.fw_idx--;
         if ( status.fw_idx < 0 )
@@ -1348,7 +1325,7 @@
       break;
 
     case grKEY( ']' ):
-      if ( status.use_custom_lcd_filter )
+      if ( status.lcd_filter < 0 )
       {
         status.fw_idx++;
         if ( status.fw_idx > 4 )
@@ -1358,10 +1335,10 @@
       break;
 
     case grKEY( '-' ):
-      if ( status.use_custom_lcd_filter )
+      if ( status.lcd_filter < 0 )
       {
         FTC_Manager_RemoveFaceID( handle->cache_manager,
-                                handle->scaler.face_id );
+                                  handle->scaler.face_id );
 
         status.filter_weights[status.fw_idx]--;
         FT_Library_SetLcdFilterWeights( handle->library,
@@ -1372,7 +1349,7 @@
 
     case grKEY( '+' ):
     case grKEY( '=' ):
-      if ( status.use_custom_lcd_filter )
+      if ( status.lcd_filter < 0 )
       {
         FTC_Manager_RemoveFaceID( handle->cache_manager,
                                   handle->scaler.face_id );
@@ -1444,19 +1421,19 @@
          status.render_mode == RENDER_MODE_WATERFALL )
       sprintf( buf, "top left charcode: U+%04X (glyph idx %d)",
                     status.topleft,
-                    FTDemo_Get_Index( handle, status.topleft ) );
+                    FTDemo_Get_Index( handle, (FT_UInt32)status.topleft ) );
     else if ( status.encoding == FT_ENCODING_NONE )
       sprintf( buf, "top left glyph idx: %d",
                     status.topleft );
     else
       sprintf( buf, "top left charcode: 0x%X (glyph idx %d)",
                     status.topleft,
-                    FTDemo_Get_Index( handle, status.topleft ) );
+                    FTDemo_Get_Index( handle, (FT_UInt32)status.topleft ) );
 
     if ( FT_HAS_GLYPH_NAMES( face ) )
     {
-      char*  p;
-      int    format_len, glyph_idx, size;
+      char*         p;
+      unsigned int  format_len, glyph_idx, size;
 
 
       size = strlen( buf );
@@ -1468,11 +1445,11 @@
 
       if ( size >= format_len + 2 )
       {
-        glyph_idx = status.topleft;
+        glyph_idx = (unsigned int)status.topleft;
         if ( status.encoding != FT_ENCODING_NONE         ||
              status.render_mode == RENDER_MODE_TEXT      ||
              status.render_mode == RENDER_MODE_WATERFALL )
-          glyph_idx = FTDemo_Get_Index( handle, status.topleft );
+          glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)status.topleft );
 
         strcpy( p, format );
         if ( FT_Get_Glyph_Name( face, glyph_idx,
@@ -1632,14 +1609,11 @@
 
     if ( handle->antialias )
     {
-      const char*  lcd_mode = NULL;
+      const char*  lcd_mode;
 
 
       switch ( handle->lcd_mode )
       {
-      case LCD_MODE_AA:
-        lcd_mode = " normal AA";
-        break;
       case LCD_MODE_LIGHT:
         lcd_mode = " light AA";
         break;
@@ -1655,6 +1629,9 @@
       case LCD_MODE_VBGR:
         lcd_mode = " LCD (vert. BGR)";
         break;
+      default:
+        handle->lcd_mode = 0;
+        lcd_mode = " normal AA";
       }
       grWriteCellString( display->bitmap, 0, (line++) * HEADER_HEIGHT,
                          lcd_mode, display->fore_color );
@@ -1700,7 +1677,8 @@
 
       else if ( !strcmp( module->clazz->module_name, "truetype" ) )
       {
-        switch ( status.tt_interpreter_version )
+        switch ( status.tt_interpreter_versions[
+                   status.tt_interpreter_version_idx] )
         {
         case TT_INTERPRETER_VERSION_35:
           hinting_engine = "v35";
@@ -1708,6 +1686,9 @@
         case TT_INTERPRETER_VERSION_38:
           hinting_engine = "v38";
           break;
+        case TT_INTERPRETER_VERSION_40:
+          hinting_engine = "v40";
+          break;
         }
       }
 
@@ -1720,6 +1701,14 @@
       }
     }
 
+    if ( handle->lcd_mode == LCD_MODE_LIGHT )
+    {
+      sprintf( buf, "warping: %s",
+                    status.warping ? "on" : "off" );
+      grWriteCellString( display->bitmap, 0, (line++) * HEADER_HEIGHT,
+                         buf, display->fore_color );
+    }
+
     line++;
 
     /* embedded bitmaps */
@@ -1741,22 +1730,26 @@
                        buf, display->fore_color );
 
     /* gamma */
-    sprintf( buf, "gamma: %.1f%s",
-                  status.gamma, status.gamma == 0.0 ? " (sRGB mode)"
-                                                    : "" );
+    if ( status.gamma == 0.0 )
+      sprintf( buf, "gamma: sRGB mode" );
+    else
+      sprintf( buf, "gamma: %.1f", status.gamma );
     grWriteCellString( display->bitmap, 0, (line++) * HEADER_HEIGHT,
                        buf, display->fore_color );
 
     line++;
 
-    /* custom LCD filtering */
-    sprintf( buf, "custom LCD: %s",
-                  status.use_custom_lcd_filter ? "on" : "off" );
+    /* LCD filtering */
+    sprintf( buf, "LCD filter: %s",
+                  status.lcd_filter == 0 ? "none" :
+                  status.lcd_filter == 1 ? "default" :
+                  status.lcd_filter == 2 ? "light" :
+                  status.lcd_filter == 3 ? "legacy" : "custom" );
     grWriteCellString( display->bitmap, 0, (line++) * HEADER_HEIGHT,
                        buf, display->fore_color );
 
-    /* LCD filter settings */
-    if ( status.use_custom_lcd_filter )
+    /* custom LCD filter settings */
+    if ( status.lcd_filter < 0 )
     {
       int             fwi = status.fw_idx;
       unsigned char*  fw  = status.filter_weights;
@@ -1897,7 +1890,7 @@
           printf( "\n" );
           exit( 0 );
         }
-        break;
+        /* break; */
 
       case 'w':
         status.width = atoi( optarg );
@@ -1931,6 +1924,11 @@
         char*  argv[] )
   {
     grEvent  event;
+    unsigned int  dflt_tt_interpreter_version;
+    int           i;
+    unsigned int  versions[3] = { TT_INTERPRETER_VERSION_35,
+                                  TT_INTERPRETER_VERSION_38,
+                                  TT_INTERPRETER_VERSION_40 };
 
 
     /* Initialize engine */
@@ -1944,9 +1942,29 @@
     FT_Property_Get( handle->library,
                      "cff",
                      "hinting-engine", &status.cff_hinting_engine );
+
+    /* collect all available versions, then set again the default */
     FT_Property_Get( handle->library,
                      "truetype",
-                     "interpreter-version", &status.tt_interpreter_version );
+                     "interpreter-version", &dflt_tt_interpreter_version );
+    for ( i = 0; i < 3; i++ )
+    {
+      error = FT_Property_Set( handle->library,
+                               "truetype",
+                               "interpreter-version", &versions[i] );
+      if ( !error )
+        status.tt_interpreter_versions[
+          status.num_tt_interpreter_versions++] = versions[i];
+      if ( versions[i] == dflt_tt_interpreter_version )
+        status.tt_interpreter_version_idx = i;
+    }
+    FT_Property_Set( handle->library,
+                     "truetype",
+                     "interpreter-version", &dflt_tt_interpreter_version );
+
+    FT_Property_Get( handle->library,
+                     "autofitter",
+                     "warping", &status.warping );
 
     handle->encoding = status.encoding;
 
@@ -1965,7 +1983,7 @@
       Fatal( "could not allocate display surface" );
 
     grSetTitle( display->surface,
-                "FreeType Glyph Viewer - press F1 for help" );
+                "FreeType Glyph Viewer - press ? for help" );
 
     status.num_fails = 0;
 
@@ -2027,7 +2045,7 @@
     FTDemo_Done( handle );
     exit( 0 );      /* for safety reasons */
 
-    return 0;       /* never reached */
+    /* return 0; */ /* never reached */
   }
 
 
